# pyre-unsafe
def resolve(path, urlconf=...): ...
def reverse(viewname, urlconf=..., args=..., kwargs=..., current_app=...): ...
def clear_url_caches(): ...
def set_script_prefix(prefix): ...
def get_script_prefix(): ...
def clear_script_prefix(): ...
def set_urlconf(urlconf_name): ...
def get_urlconf(default=...): ...
def is_valid_path(path, urlconf=...): ...
def translate_url(url, lang_code): ...
